
# THIS FILE IS GENERATED FROM OASYS SETUP.PY
short_version = '1.0.11'
version = '1.0.11'
full_version = '1.0.11'
git_revision = 'f06da4ec995f1102aeb467c95b8fb989dbf47c06'
release = True

if not release:
    version = full_version
    short_version += ".dev"
